"""
benchmark asammdf vs mdfreader
"""

import argparse
from io import StringIO
import multiprocessing
import os
import platform
import sys
import traceback

from mdfreader import __version__ as mdfreader_version
from mdfreader import Mdf as MDFreader
import numpy as np
import psutil

from asammdf import __version__ as asammdf_version
from asammdf import MDF, Signal
import asammdf.blocks.v2_v3_blocks as v3b
import asammdf.blocks.v2_v3_constants as v3c
import asammdf.blocks.v4_blocks as v4b
import asammdf.blocks.v4_constants as v4c

try:
    import resource
except ImportError:
    pass

PYVERSION = sys.version_info[0]

if PYVERSION > 2:
    from time import perf_counter
else:
    from time import clock as perf_counter


class MyList(list):
    """list that prints the items that are appended or extended"""

    def append(self, item):
        """append item and print it to stdout"""
        print(item)
        super().append(item)

    def extend(self, items):
        """extend items and print them to stdout
        using the new line separator
        """
        print("\n".join(items))
        super().extend(items)


class Timer:
    """measures the RAM usage and elased time. The information is saved in
    the output attribute and any Exception text is saved in the error attribute

    Parameters
    ----------
    topic : str
        timer title; only used if Exceptions are raised during execution
    message : str
        execution item description
    fmt : str
        output fmt; can be "rst" (rstructured text) or "md" (markdown)
    """

    def __init__(self, topic, message, fmt="rst"):
        self.topic = topic
        self.message = message
        self.output = ""
        self.error = ""
        self.fmt = fmt
        self.start = None

    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, type_, value, tracebackobj):
        elapsed_time = int((perf_counter() - self.start) * 1000)
        process = psutil.Process(os.getpid())

        if platform.system() == "Windows":
            ram_usage = int(process.memory_info().peak_wset / 1024 / 1024)
        else:
            ram_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            ram_usage = int(ram_usage / 1024)

        if tracebackobj:
            info = StringIO()
            traceback.print_tb(tracebackobj, None, info)
            info.seek(0)
            info = info.read()
            self.error = f"{self.topic} : {self.message}\n{type_}\t \n{value}{info}"
            if self.fmt == "rst":
                self.output = "{:<50} {:>9} {:>8}".format(self.message, "0*", "0*")
            elif self.fmt == "md":
                self.output = "|{:<50}|{:>9}|{:>8}|".format(self.message, "0*", "0*")
        else:
            if self.fmt == "rst":
                self.output = f"{self.message:<50} {elapsed_time:>9} {ram_usage:>8}"
            elif self.fmt == "md":
                self.output = f"|{self.message:<50}|{elapsed_time:>9}|{ram_usage:>8}|"

        return True


def generate_test_files(version="4.10"):
    cycles = 3000
    channels_count = 2000
    mdf = MDF(version=version)

    if version <= "3.30":
        filename = r"test.mdf"
    else:
        filename = r"test.mf4"

    if os.path.exists(filename):
        return filename

    t = np.arange(cycles, dtype=np.float64)

    cls = v4b.ChannelConversion if version >= "4.00" else v3b.ChannelConversion

    # no conversion
    sigs = []
    for i in range(channels_count):
        sig = Signal(
            np.ones(cycles, dtype=np.uint64) * i,
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=None,
            comment=f"Unsigned int 16bit channel {i}",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # linear
    sigs = []
    for i in range(channels_count):
        conversion = {
            "conversion_type": v4c.CONVERSION_TYPE_LIN if version >= "4.00" else v3c.CONVERSION_TYPE_LINEAR,
            "a": float(i),
            "b": -0.5,
        }
        sig = Signal(
            np.ones(cycles, dtype=np.int64),
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=cls(**conversion),
            comment=f"Signed 16bit channel {i} with linear conversion",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # algebraic
    sigs = []
    for i in range(channels_count):
        conversion = {
            "conversion_type": v4c.CONVERSION_TYPE_ALG if version >= "4.00" else v3c.CONVERSION_TYPE_FORMULA,
            "formula": f"{i} * sin(X)",
        }
        sig = Signal(
            np.arange(cycles, dtype=np.int32) / 100.0,
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=cls(**conversion),
            comment=f"Sinus channel {i} with algebraic conversion",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # rational
    sigs = []
    for i in range(channels_count):
        conversion = {
            "conversion_type": v4c.CONVERSION_TYPE_RAT if version >= "4.00" else v3c.CONVERSION_TYPE_RAT,
            "P1": 0,
            "P2": i,
            "P3": -0.5,
            "P4": 0,
            "P5": 0,
            "P6": 1,
        }
        sig = Signal(
            np.ones(cycles, dtype=np.int64),
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=cls(**conversion),
            comment=f"Channel {i} with rational conversion",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # string
    sigs = []
    for i in range(channels_count):
        sig = [f"Channel {i} sample {j}".encode("ascii") for j in range(cycles)]
        sig = Signal(
            np.array(sig),
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            comment=f"String channel {i}",
            raw=True,
            encoding="utf-8",
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # byte array
    sigs = []
    ones = np.ones(cycles, dtype=np.dtype("(8,)u1"))
    for i in range(channels_count):
        sig = Signal(
            ones * (i % 255),
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            comment=f"Byte array channel {i}",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # value to text
    sigs = []
    ones = np.ones(cycles, dtype=np.uint64)
    conversion = {
        "raw": np.arange(255, dtype=np.float64),
        "phys": np.array([f"Value {i}".encode("ascii") for i in range(255)]),
        "conversion_type": v4c.CONVERSION_TYPE_TABX if version >= "4.00" else v3c.CONVERSION_TYPE_TABX,
        "links_nr": 260,
        "ref_param_nr": 255,
    }

    for i in range(255):
        conversion[f"val_{i}"] = conversion[f"param_val_{i}"] = conversion["raw"][i]
        conversion[f"text_{i}"] = conversion["phys"][i]
    conversion[f"text_{255}"] = "Default"

    for i in range(channels_count):
        sig = Signal(
            ones * i,
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            comment=f"Value to text channel {i}",
            conversion=cls(**conversion),
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    mdf.save(filename, overwrite=True)


def open_mdf3(output, fmt):
    with Timer("Open file", f"asammdf {asammdf_version} mdfv3", fmt) as timer:
        MDF(r"test.mdf")
    output.send([timer.output, timer.error])


def open_mdf4(output, fmt):
    with Timer("Open file", f"asammdf {asammdf_version} mdfv4", fmt) as timer:
        MDF(r"test.mf4")
    output.send([timer.output, timer.error])


def open_mdf4_column(output, fmt):
    with Timer("Open file", f"asammdf {asammdf_version} column mdfv4", fmt) as timer:
        MDF(r"test_column.mf4")
    output.send([timer.output, timer.error])


def save_mdf3(output, fmt):
    x = MDF(r"test.mdf")
    with Timer("Save file", f"asammdf {asammdf_version} mdfv3", fmt) as timer:
        x.save(r"x.mdf", overwrite=True)
    output.send([timer.output, timer.error])


def save_mdf4(output, fmt):
    x = MDF(r"test.mf4")
    with Timer("Save file", f"asammdf {asammdf_version} mdfv4", fmt) as timer:
        x.save(r"x.mf4", overwrite=True)
    output.send([timer.output, timer.error])


def save_mdf4_column(output, fmt):
    x = MDF(r"test_column.mf4")
    with Timer("Save file", f"asammdf {asammdf_version} mdfv4 column", fmt) as timer:
        x.save(r"x.mf4", overwrite=True)
    output.send([timer.output, timer.error])


def get_all_mdf3(output, fmt):
    x = MDF(r"test.mdf")
    with Timer("Get all channels", f"asammdf {asammdf_version} mdfv3", fmt) as timer:
        for i, gp in enumerate(x.groups):
            for j in range(len(gp["channels"])):
                x.get(group=i, index=j, samples_only=True)
    output.send([timer.output, timer.error])


def get_all_mdf4(output, fmt):
    x = MDF(r"test.mf4")
    with Timer("Get all channels", f"asammdf {asammdf_version} mdfv4", fmt) as timer:
        t = perf_counter()
        counter = 0
        to_break = False
        for i, gp in enumerate(x.groups):
            if to_break:
                break
            for j in range(len(gp["channels"])):
                t2 = perf_counter()
                if t2 - t > 60:
                    timer.message += f" {counter / (t2 - t)}/s"
                    to_break = True
                    break
                x.get(group=i, index=j, samples_only=True)
                counter += 1
    output.send([timer.output, timer.error])


def get_all_mdf4_column(output, fmt):
    x = MDF(r"test_column.mf4")
    with Timer("Get all channels", f"asammdf {asammdf_version} column mdfv4", fmt) as timer:
        t = perf_counter()
        counter = 0
        to_break = False
        for i, gp in enumerate(x.groups):
            if to_break:
                break
            for j in range(len(gp["channels"])):
                t2 = perf_counter()
                if t2 - t > 60:
                    timer.message += f" {counter / (t2 - t)}/s"
                    to_break = True
                    break
                x.get(group=i, index=j, samples_only=False)
                counter += 1
    output.send([timer.output, timer.error])


def convert_v3_v4(output, fmt):
    with MDF(r"test.mdf") as x:
        with Timer("Convert file", f"asammdf {asammdf_version} v3 to v4", fmt) as timer:
            x.convert("4.10")
    output.send([timer.output, timer.error])


def convert_v4_v410(output, fmt):
    with MDF(r"test.mf4") as x:
        with Timer("Convert file", f"asammdf {asammdf_version} v4 to v410", fmt) as timer:
            y = x.convert("4.10")
            y.close()
    output.send([timer.output, timer.error])


def convert_v4_v420(output, fmt):
    with MDF(r"test.mf4") as x:
        with Timer("Convert file", f"asammdf {asammdf_version} v4 to v420", fmt) as timer:
            y = x.convert("4.20")
            y.close()
    output.send([timer.output, timer.error])


def merge_v3(output, fmt):
    files = [r"test.mdf"] * 3
    with Timer("Merge 3 files", f"asammdf {asammdf_version} v3", fmt) as timer:
        MDF.concatenate(files, outversion="3.30")
    output.send([timer.output, timer.error])


def merge_v4(output, fmt):
    files = [r"test.mf4"] * 3

    with Timer("Merge 3 files", f"asammdf {asammdf_version} v4", fmt) as timer:
        MDF.concatenate(files, outversion="4.10")
    output.send([timer.output, timer.error])


#
# mdfreader
#


def open_reader3(output, fmt):
    with Timer("Open file", f"mdfreader {mdfreader_version} mdfv3", fmt) as timer:
        MDFreader(r"test.mdf")
    output.send([timer.output, timer.error])


def open_reader3_nodata(output, fmt):
    with Timer("Open file", f"mdfreader {mdfreader_version} no_data_loading mdfv3", fmt) as timer:
        MDFreader(r"test.mdf", no_data_loading=True)
    output.send([timer.output, timer.error])


def open_reader3_compression(output, fmt):
    with Timer("Open file", f"mdfreader {mdfreader_version} compress mdfv3", fmt) as timer:
        MDFreader(r"test.mdf", compression="blosc")
    output.send([timer.output, timer.error])


def open_reader4(output, fmt):
    with Timer("Open file", f"mdfreader {mdfreader_version} mdfv4", fmt) as timer:
        MDFreader(r"test.mf4")
    output.send([timer.output, timer.error])


def open_reader4_nodata(output, fmt):
    with Timer("Open file", f"mdfreader {mdfreader_version} no_data_loading mdfv4", fmt) as timer:
        MDFreader(r"test.mf4", no_data_loading=True)
    output.send([timer.output, timer.error])


def open_reader4_compression(output, fmt):
    with Timer("Open file", f"mdfreader {mdfreader_version} compress mdfv4", fmt) as timer:
        MDFreader(r"test.mf4", compression="blosc")
    output.send([timer.output, timer.error])


def save_reader3(output, fmt):
    x = MDFreader(r"test.mdf")
    with Timer("Save file", f"mdfreader {mdfreader_version} mdfv3", fmt) as timer:
        x.write(r"x.mdf")
    output.send([timer.output, timer.error])


def save_reader3_nodata(output, fmt):
    x = MDFreader(r"test.mdf", no_data_loading=True)
    with Timer("Save file", f"mdfreader {mdfreader_version} no_data_loading mdfv3", fmt) as timer:
        x.write(r"x.mdf")
    output.send([timer.output, timer.error])


def save_reader3_compression(output, fmt):
    with Timer("Save file", f"mdfreader {mdfreader_version} compress mdfv3", fmt) as outer_timer:
        x = MDFreader(r"test.mdf", compression="blosc")
        with Timer("Save file", f"mdfreader {mdfreader_version} compress mdfv3", fmt) as timer:
            x.write(r"x.mdf")
        output.send([timer.output, timer.error])
    if outer_timer.error:
        output.send([timer.output, timer.error])


def save_reader4(output, fmt):
    x = MDFreader(r"test.mf4")
    with Timer("Save file", f"mdfreader {mdfreader_version} mdfv4", fmt) as timer:
        x.write(r"x.mf4")
    output.send([timer.output, timer.error])


def save_reader4_nodata(output, fmt):
    x = MDFreader(r"test.mf4", no_data_loading=True)
    with Timer("Save file", f"mdfreader {mdfreader_version} no_data_loading mdfv4", fmt) as timer:
        x.write(r"x.mf4")
    output.send([timer.output, timer.error])


def save_reader4_compression(output, fmt):
    x = MDFreader(r"test.mf4", compression="blosc")
    with Timer("Save file", f"mdfreader {mdfreader_version} compress mdfv4", fmt) as timer:
        x.write(r"x.mf4")
    output.send([timer.output, timer.error])


def get_all_reader3(output, fmt):
    x = MDFreader(r"test.mdf")
    with Timer("Get all channels", f"mdfreader {mdfreader_version} mdfv3", fmt) as timer:
        for s in x:
            x.get_channel_data(s)
    output.send([timer.output, timer.error])


def get_all_reader3_nodata(output, fmt):
    x = MDFreader(r"test.mdf", no_data_loading=True)
    with Timer("Get all channels", f"mdfreader {mdfreader_version} nodata mdfv3", fmt) as timer:
        for s in x:
            x.get_channel_data(s)
    output.send([timer.output, timer.error])


def get_all_reader3_compression(output, fmt):
    x = MDFreader(r"test.mdf", compression="blosc")
    with Timer("Get all channels", f"mdfreader {mdfreader_version} compress mdfv3", fmt) as timer:
        for s in x:
            x.get_channel_data(s)

        with open("D:\\TMP\\f.txt", "w") as f:
            f.write("OK")
    output.send([timer.output, timer.error])


def get_all_reader4(output, fmt):
    x = MDFreader(r"test.mf4")
    with Timer("Get all channels", f"mdfreader {mdfreader_version} mdfv4", fmt) as timer:
        t = perf_counter()
        counter = 0
        to_break = False
        for s in x:
            t2 = perf_counter()
            if t2 - t > 60:
                timer.message += f" {counter / (t2 - t)}/s"
                to_break = True
                break
            x.get_channel_data(s)
            counter += 1
    output.send([timer.output, timer.error])


def get_all_reader4_nodata(output, fmt):
    x = MDFreader(r"test.mf4", no_data_loading=True)
    with Timer("Get all channels", f"mdfreader {mdfreader_version} nodata mdfv4", fmt) as timer:
        t = perf_counter()
        counter = 0
        to_break = False
        for s in x:
            t2 = perf_counter()
            if t2 - t > 60:
                timer.message += f" {counter / (t2 - t)}/s"
                to_break = True
                break
            x.get_channel_data(s)
            counter += 1
    output.send([timer.output, timer.error])


def get_all_reader4_compression(output, fmt):
    x = MDFreader(r"test.mf4", compression="blosc")
    with Timer("Get all channels", f"mdfreader {mdfreader_version} compress mdfv4", fmt) as timer:
        t = perf_counter()
        counter = 0
        to_break = False
        for s in x:
            t2 = perf_counter()
            if t2 - t > 60:
                timer.message += f" {counter / (t2 - t)}/s"
                to_break = True
                break
            x.get_channel_data(s)
            counter += 1
    output.send([timer.output, timer.error])


def merge_reader_v3(output, fmt):
    files = [r"test.mdf"] * 3
    with Timer("Merge 3 files", f"mdfreader {mdfreader_version} v3", fmt) as timer:
        x1 = MDFreader(files[0])
        x1.resample(0.01)
        x2 = MDFreader(files[1])
        x2.resample(0.01)
        x1.merge_mdf(x2)
        x2 = MDFreader(files[2])
        x2.resample(0.01)
        x1.merge_mdf(x2)
    output.send([timer.output, timer.error])


def merge_reader_v3_compress(output, fmt):
    files = [r"test.mdf"] * 3
    with Timer("Merge 3 files", f"mdfreader {mdfreader_version} compress v3", fmt) as timer:
        x1 = MDFreader(files[0], compression="blosc")
        x1.resample(0.01)
        x2 = MDFreader(files[1], compression="blosc")
        x2.resample(0.01)
        x1.merge_mdf(x2)
        x2 = MDFreader(files[2], compression="blosc")
        x2.resample(0.01)
        x1.merge_mdf(x2)
    output.send([timer.output, timer.error])


def merge_reader_v3_nodata(output, fmt):
    files = [r"test.mdf"] * 3
    with Timer("Merge 3 files", f"mdfreader {mdfreader_version} nodata v3", fmt) as timer:
        x1 = MDFreader(files[0], no_data_loading=True)
        x1.resample(0.01)
        x2 = MDFreader(files[1], no_data_loading=True)
        x2.resample(0.01)
        x1.merge_mdf(x2)
        x2 = MDFreader(files[2], no_data_loading=True)
        x2.resample(0.01)
        x1.merge_mdf(x2)
    output.send([timer.output, timer.error])


def merge_reader_v4(output, fmt):
    files = [r"test.mf4"] * 3

    with Timer("Merge 3 files", f"mdfreader {mdfreader_version} v4", fmt) as timer:
        x1 = MDFreader(files[0])
        x1.resample(0.01)
        x2 = MDFreader(files[1])
        x2.resample(0.01)
        x1.merge_mdf(x2)
        x2 = MDFreader(files[2])
        x2.resample(0.01)
        x1.merge_mdf(x2)

    output.send([timer.output, timer.error])


def merge_reader_v4_compress(output, fmt):
    files = [r"test.mf4"] * 3
    with Timer("Merge 3 files", f"mdfreader {mdfreader_version} compress v4", fmt) as timer:
        x1 = MDFreader(files[0], compression="blosc")
        x1.resample(0.01)
        x2 = MDFreader(files[1], compression="blosc")
        x2.resample(0.01)
        x1.merge_mdf(x2)
        x2 = MDFreader(files[2], compression="blosc")
        x2.resample(0.01)
        x1.merge_mdf(x2)

    output.send([timer.output, timer.error])


def merge_reader_v4_nodata(output, fmt):
    files = [r"test.mf4"] * 3
    with Timer("Merge 3 files", f"mdfreader {mdfreader_version} nodata v4", fmt) as timer:
        x1 = MDFreader(files[0], no_data_loading=True)
        x1.resample(0.01)
        x2 = MDFreader(files[1], no_data_loading=True)
        x2.resample(0.01)
        x1.merge_mdf(x2)
        x2 = MDFreader(files[2], no_data_loading=True)
        x2.resample(0.01)
        x1.merge_mdf(x2)

    output.send([timer.output, timer.error])


#
# utility functions
#


def filter_asam(output, fmt):
    with Timer("Filter file", f"asammdf {asammdf_version} mdfv4", fmt) as timer:
        x = MDF(r"test.mf4").filter([(None, i, int(f"{j}5")) for i in range(10, 20) for j in range(1, 20)])
        t = perf_counter()
        counter = 0
        to_break = False
        for i, gp in enumerate(x.groups):
            if to_break:
                break
            for j in range(len(gp["channels"])):
                t2 = perf_counter()
                if t2 - t > 60:
                    timer.message += f" {counter / (t2 - t)}/s"
                    to_break = True
                    break
                x.get(group=i, index=j, samples_only=True)
                counter += 1
    output.send([timer.output, timer.error])


def filter_reader4(output, fmt):
    with Timer("Filter file", f"mdfreader {mdfreader_version} mdfv4", fmt) as timer:
        x = MDFreader(
            r"test.mf4",
            channel_list=[f"Channel_{i}_{j}5" for i in range(10) for j in range(1, 20)],
        )
        t = perf_counter()
        counter = 0
        for s in x:
            t2 = perf_counter()
            if t2 - t > 60:
                timer.message += f" {counter / (t2 - t)}/s"
                break
            x.get_channel_data(s)
            counter += 1
    output.send([timer.output, timer.error])


def filter_reader4_compression(output, fmt):
    with Timer("Filter file", f"mdfreader {mdfreader_version} compression mdfv4", fmt) as timer:
        x = MDFreader(
            r"test.mf4",
            compression="blosc",
            channel_list=[f"Channel_{i}_{j}5" for i in range(10) for j in range(1, 20)],
        )
        t = perf_counter()
        counter = 0
        for s in x:
            t2 = perf_counter()
            if t2 - t > 60:
                timer.message += f" {counter / (t2 - t)}/s"
                break
            x.get_channel_data(s)
            counter += 1
    output.send([timer.output, timer.error])


def filter_reader4_nodata(output, fmt):
    with Timer("Filter file", f"mdfreader {mdfreader_version} nodata mdfv4", fmt) as timer:
        x = MDFreader(
            r"test.mf4",
            no_data_loading=True,
            channel_list=[f"Channel_{i}_{j}5" for i in range(10) for j in range(1, 20)],
        )
        t = perf_counter()
        counter = 0
        for s in x:
            t2 = perf_counter()
            if t2 - t > 60:
                timer.message += f" {counter / (t2 - t)}/s"
                break
            x.get_channel_data(s)
            counter += 1
    output.send([timer.output, timer.error])


def cut_asam(output, fmt):
    x = MDF(r"test.mf4")
    t = x.get_master(0)
    start, stop = 0.2 * (t[-1] - t[0]) + t[0], 0.8 * (t[-1] - t[0]) + t[0]
    with Timer("Cut file", f"asammdf {asammdf_version} mdfv4", fmt) as timer:
        x = x.cut(start=start, stop=stop)

    output.send([timer.output, timer.error])


def cut_reader4(output, fmt):
    x = MDFreader(r"test.mf4")
    t = x.get_channel_data(list(x.masterChannelList)[0])
    begin, end = 0.2 * (t[-1] - t[0]) + t[0], 0.8 * (t[-1] - t[0]) + t[0]
    with Timer("Cut file", f"mdfreader {mdfreader_version} mdfv4", fmt) as timer:
        x.cut(begin=begin, end=end)
    output.send([timer.output, timer.error])


def cut_reader4_compression(output, fmt):
    x = MDFreader(r"test.mf4", compression="blosc")
    t = x.get_channel_data(list(x.masterChannelList)[0])
    begin, end = 0.2 * (t[-1] - t[0]) + t[0], 0.8 * (t[-1] - t[0]) + t[0]
    with Timer("Cut file", f"mdfreader {mdfreader_version} compression mdfv4", fmt) as timer:
        x.cut(begin=begin, end=end)
    output.send([timer.output, timer.error])


def cut_reader4_nodata(output, fmt):
    x = MDFreader(r"test.mf4", no_data_loading=True)
    t = x.get_channel_data(list(x.masterChannelList)[0])
    begin, end = 0.2 * (t[-1] - t[0]) + t[0], 0.8 * (t[-1] - t[0]) + t[0]
    with Timer("Cut file", f"mdfreader {mdfreader_version} nodata mdfv4", fmt) as timer:
        x.cut(begin=begin, end=end)
    output.send([timer.output, timer.error])


def table_header(topic, fmt="rst"):
    output = []
    if fmt == "rst":
        result = "{:<50} {:>9} {:>8}".format(topic, "Time [ms]", "RAM [MB]")
        output.append("")
        output.append("{} {} {}".format("=" * 50, "=" * 9, "=" * 8))
        output.append(result)
        output.append("{} {} {}".format("=" * 50, "=" * 9, "=" * 8))
    elif fmt == "md":
        result = "|{:<50}|{:>9}|{:>8}|".format(topic, "Time [ms]", "RAM [MB]")
        output.append("")
        output.append(result)
        output.append("|{}|{}|{}|".format("-" * 50, "-" * 9, "-" * 8))
    return output


def table_end(fmt="rst"):
    if fmt == "rst":
        return ["{} {} {}".format("=" * 50, "=" * 9, "=" * 8), ""]
    elif fmt == "md":
        return [
            "",
        ]


def main(text_output, fmt):
    if os.path.dirname(__file__):
        os.chdir(os.path.dirname(__file__))
    for version in ("3.30", "4.10"):
        generate_test_files(version)

    mdf = MDF("test.mdf", "minimum")
    v3_size = os.path.getsize("test.mdf") // 1024 // 1024
    v3_groups = len(mdf.groups)
    v3_channels = sum(len(gp["channels"]) for gp in mdf.groups)
    v3_version = mdf.version

    mdf = MDF("test.mf4", "minimum")
    mdf.get_master(0)
    v4_size = os.path.getsize("test.mf4") // 1024 // 1024
    v4_groups = len(mdf.groups)
    v4_channels = sum(len(gp["channels"]) for gp in mdf.groups)
    v4_version = mdf.version

    listen, send = multiprocessing.Pipe()
    output = MyList()
    errors = []

    installed_ram = round(psutil.virtual_memory().total / 1024 / 1024 / 1024)

    output.append("\n\nBenchmark environment\n")
    output.append(f"* {sys.version}")
    output.append(f"* {platform.platform()}")
    output.append(f"* {platform.processor()}")
    output.append(f"* numpy {np.__version__}")
    output.append(f"* {installed_ram}GB installed RAM\n")
    output.append("Notations used in the results\n")
    output.append("* compress = mdfreader mdf object created with " "compression=blosc")
    output.append("* nodata = mdfreader mdf object read with " "no_data_loading=True")
    output.append("\nFiles used for benchmark:\n")
    output.append(f"* mdf version {v3_version}")
    output.append(f"    * {v3_size} MB file size")
    output.append(f"    * {v3_groups} groups")
    output.append(f"    * {v3_channels} channels")
    output.append(f"* mdf version {v4_version}")
    output.append(f"    * {v4_size} MB file size")
    output.append(f"    * {v4_groups} groups")
    output.append(f"    * {v4_channels} channels\n\n")

    OPEN, SAVE, GET, CONVERT, MERGE, FILTER, CUT = 1, 1, 1, 1, 1, 1, 1

    tests = (
        open_mdf3,
        # open_reader3,
        # open_reader3_nodata,
        # open_reader3_compression,
        open_mdf4,
        # open_mdf4_column,
        # open_reader4,
        # open_reader4_nodata,
        # open_reader4_compression,
    )

    if tests and OPEN:
        output.extend(table_header("Open file", fmt))
        for func in tests:
            thr = multiprocessing.Process(target=func, args=(send, fmt))
            thr.start()
            thr.join()
            result, err = listen.recv()
            output.append(result)
            errors.append(err)
        output.extend(table_end(fmt))

    tests = (
        save_mdf3,
        # save_reader3,
        # save_reader3_nodata,
        # save_reader3_compression,
        save_mdf4,
        # save_mdf4_column,
        # save_reader4,
        # save_reader4_nodata,
        # save_reader4_compression,
    )

    if tests and SAVE:
        output.extend(table_header("Save file", fmt))
        for func in tests:
            thr = multiprocessing.Process(target=func, args=(send, fmt))
            thr.start()
            thr.join()
            result, err = listen.recv()
            output.append(result)
            errors.append(err)
        output.extend(table_end(fmt))

    tests = (
        get_all_mdf3,
        # get_all_reader3,
        # get_all_reader3_nodata,
        # get_all_reader3_compression,
        get_all_mdf4,
        # get_all_mdf4_column,
        # get_all_reader4,
        # get_all_reader4_nodata,
        # get_all_reader4_compression,
    )

    if tests and GET:
        output.extend(table_header("Get all channels (36424 calls)", fmt))
        for func in tests:
            thr = multiprocessing.Process(target=func, args=(send, fmt))
            thr.start()
            thr.join()
            result, err = listen.recv()
            output.append(result)
            errors.append(err)
        output.extend(table_end(fmt))

    tests = (
        convert_v3_v4,
        convert_v4_v410,
        convert_v4_v420,
    )

    if tests and CONVERT:
        output.extend(table_header("Convert file", fmt))
        for func in tests:
            thr = multiprocessing.Process(target=func, args=(send, fmt))
            thr.start()
            thr.join()
            result, err = listen.recv()
            output.append(result)
            errors.append(err)
        output.extend(table_end(fmt))

    tests = (
        merge_v3,
        # merge_reader_v3,
        # merge_reader_v3_nodata,
        # merge_reader_v3_compress,
        merge_v4,
        # merge_reader_v4,
        # merge_reader_v4_nodata,
        # merge_reader_v4_compress,
    )

    if tests and MERGE:
        output.extend(table_header("Merge 3 files", fmt))
        for func in tests:
            thr = multiprocessing.Process(target=func, args=(send, fmt))
            thr.start()
            thr.join()
            result, err = listen.recv()
            output.append(result)
            errors.append(err)
        output.extend(table_end(fmt))

    tests = (
        filter_asam,
        # filter_reader4,
        # filter_reader4_compression,
        # filter_reader4_nodata,
    )

    if tests and FILTER:
        output.extend(table_header("Filter 200 channels", fmt))
        for func in tests:
            thr = multiprocessing.Process(target=func, args=(send, fmt))
            thr.start()
            thr.join()
            result, err = listen.recv()
            output.append(result)
            errors.append(err)
        output.extend(table_end(fmt))

    tests = (
        cut_asam,
        # cut_reader4,
        # cut_reader4_compression,
        # cut_reader4_nodata,
    )

    if tests and CUT:
        output.extend(table_header("Cut file from 20% to 80%", fmt))
        for func in tests:
            thr = multiprocessing.Process(target=func, args=(send, fmt))
            thr.start()
            thr.join()
            result, err = listen.recv()
            output.append(result)
            errors.append(err)
        output.extend(table_end(fmt))

    errors = [err for err in errors if err]
    if errors:
        print("\n\nERRORS\n", "\n".join(errors))

    if text_output:
        arch = "x86" if platform.architecture()[0] == "32bit" else "x64"
        file = f"{arch}_asammdf_{asammdf_version}_mdfreader_{mdfreader_version}.{fmt}"
        with open(file, "w") as out:
            out.write("\n".join(output))

    for file in ("x.mdf", "x.mf4"):
        if PYVERSION >= 3:
            try:
                os.remove(file)
            except FileNotFoundError:
                pass
        else:
            try:
                os.remove(file)
            except OSError:
                pass


def _cmd_line_parser():
    """
    return a command line parser. It is used when generating the documentation
    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--path",
        help=("path to test files, " "if not provided the script folder is used"),
    )
    parser.add_argument(
        "--text_output",
        action="store_true",
        help="option to save the results to text file",
    )
    parser.add_argument(
        "--format",
        default="rst",
        nargs="?",
        choices=["rst", "md"],
        help="text formatting",
    )

    return parser


if __name__ == "__main__":
    cmd_parser = _cmd_line_parser()
    args = cmd_parser.parse_args(sys.argv[1:])

    main(args.text_output, args.format)
#
#    x = MDF(r"test_column.mf4")
#    with Timer("Get all channels", f"asammdf {asammdf_version} column mdfv4", "rst") as timer:
#        t = perf_counter()
#        counter = 0
#        to_break = False
#        for i, gp in enumerate(x.groups):
#            if to_break:
#                break
#            for j in range(len(gp["channels"])):
#                t2 = perf_counter()
#                if t2 - t > 60:
#                    timer.message += " {}/s".format(counter / (t2 - t))
#                    to_break = True
#                    break
#                x.get(group=i, index=j, samples_only=True)
#                counter += 1
#
#    print(timer.output, timer.error)



#!/usr/bin/env python
import datetime
from math import ceil
from pathlib import Path
from random import randint
import unittest
from unittest import mock
import urllib
import urllib.request
from zipfile import ZipFile

from can.io import ASCReader
import numpy as np
import pandas as pd
from PySide6 import QtCore, QtTest, QtWidgets
import scipy

from test.asammdf.gui.test_base import OpenHDF5, OpenMDF
from test.asammdf.gui.widgets.test_BaseBatchWidget import TestBatchWidget

# Note: If it's possible and make sense, use self.subTests
# to avoid initializing widgets multiple times and consume time.


class TestPushButtonScrambleTexts(TestBatchWidget):
    def test_ScrambleTexts(self):
        """
        Events:
            - Open 'BatchWidget' with valid measurement.
            - Go to Tab: "Modify & Export": Index 1
            - Press PushButton "Scramble texts"

        Evaluate:
            - New file is created
            - No channel from first file is found in 2nd file (scrambled file)
        """
        # Prepare workspace
        self.copy_mdf_files_to_workspace()

        test_file = Path(self.test_workspace, self.default_test_file)
        scrambled_filepath = Path(self.test_workspace, self.default_test_file.replace(".", ".scrambled."))

        # Evaluate
        self.assertFalse(scrambled_filepath.exists())

        # Setup
        self.setUpBatchWidget(measurement_files=[str(test_file)])

        channels = []
        iterator = QtWidgets.QTreeWidgetItemIterator(self.widget.filter_tree)
        while item := iterator.value():
            iterator += 1
            channels.append(item.name)

        # Go to Tab: "Modify & Export": Index 1
        self.widget.aspects.setCurrentIndex(self.modify_aspect)

        # Events
        QtTest.QTest.mouseClick(self.widget.scramble_btn, QtCore.Qt.MouseButton.LeftButton)

        # Wait for Thread to finish
        self.processEvents(1)

        # Evaluate
        self.assertTrue(scrambled_filepath.exists())

        # get saved file as MDF
        with OpenMDF(scrambled_filepath) as mdf_file:
            # Evaluate file
            for name in channels:
                self.assertNotIn(name, mdf_file.channels_db)


class TestPushButtonApply(TestBatchWidget):
    def setUp(self):
        """
        Events
            - Open 'BatchWidget' with valid measurement.
            - Go to Tab: "Modify & Export": Index 1
            - Set "channel_view" to "Natural sort"
            - Select some channels
            - Set test workspace directory as output folder

        Evaluate
            Output folder text is changed to test workspace directory

        """
        super().setUp()
        self.measurement_file = str(Path(self.resource, self.default_test_file))

        self.setUpBatchWidget(measurement_files=None)

        # Go to Tab: "Modify & Export": Index 1
        self.widget.aspects.setCurrentIndex(self.modify_aspect)

        self.widget.filter_view.setCurrentText("Natural sort")
        self.processEvents(0.1)

        # set test_workspace folder as output folder
        with mock.patch(
            "asammdf.gui.widgets.batch.QtWidgets.QFileDialog.getExistingDirectory"
        ) as mo_getExistingDirectory:
            mo_getExistingDirectory.return_value = self.test_workspace
            QtTest.QTest.mouseClick(self.widget.modify_output_folder_btn, QtCore.Qt.MouseButton.LeftButton)

        self.assertEqual(self.widget.modify_output_folder.text().strip(), self.test_workspace)

        self.tested_btn = self.widget.apply_btn

    def generic_setup(self, name: str = "MDF", check: bool = True):  # Parquet is removed?
        """
        Add default test file to files list
        Check checkboxes for random channels.
        Set check state for output options widget checkboxes .

        Parameters
        ----------
        name: name of output options widget
        check: check state for output options widget checkboxes
        """
        self.widget.files_list.addItems([str(Path(self.resource, self.default_test_file))])
        self.select_random_channels()

        for q_widget in self.widget.output_format.children():
            if name in q_widget.objectName():
                break
        else:
            unittest.skip(f"{name} output option widget is not defined")

        self.toggle_checkboxes(widget=self.widget.output_options, check=check)

        self.processEvents(0.01)

    def select_random_channels(self):
        count = self.widget.filter_tree.topLevelItemCount()
        self.select_channels(set(np.random.randint(0, count, size=randint(int(count / 4), int(count / 2)))))
        self.processEvents(0.01)

    def get_mdf_from_git(self) -> Path:
        url = "https://github.com/danielhrisca/asammdf/files/4328945/OBD2-DBC-MDF4.zip"
        urllib.request.urlretrieve(url, "test.zip")
        ZipFile(r"test.zip").extractall(self.test_workspace)
        Path("test.zip").unlink()
        temp_dir = Path(self.test_workspace)

        # Get mdf file path
        return [input_file for input_file in temp_dir.iterdir() if input_file.suffix == ".mf4"][0]

    @staticmethod
    def interpolated_timestamps(mdf, channels: set):
        all_ts = []
        for group_index, gp in enumerate(mdf.groups):
            if gp.channel_group.cycles_nr and (channels - {ch.name for ch in gp.channels}) != channels:
                all_ts.append(mdf.get_master(group_index))
        if all_ts:
            return np.unique(np.concatenate(all_ts))
        else:
            return np.array([], dtype="f8")

    def test_output_format_MDF(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Cut and resample features will be tested in separate tests.
        Events:
            - Ensure that output format is MDF
            - Press PushButton Apply.

        Evaluate:
            - File was created.
            - Ensure that output file has only selected channels
        """
        # Expected result
        saved_file = Path(self.test_workspace, self.default_test_file)

        # Ensure output format
        name = "MDF"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(name, check=False)

        # Event
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(saved_file.exists())

        # get saved file as MDF
        with OpenMDF(saved_file) as mdf_file:
            # Evaluate saved file
            for channel in self.selected_channels:
                self.assertIn(channel, mdf_file.channels_db)

    def test_output_format_ASC(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Events:
            - Ensure that output format is ASC
            - Press PushButton Apply.

        Evaluate:
            - File was created.
            - Ensure that output file has only selected channels data
        """
        # Get test file from git archive
        mdf_path = self.get_mdf_from_git()
        # Add extracted file to files list
        self.widget.files_list.addItems([str(mdf_path)])

        # Expected results
        asc_path = Path.with_suffix(Path(self.test_workspace, mdf_path.stem), ".asc")

        # Ensure output format
        self.widget.output_format.setCurrentText("ASC")

        # Event
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(asc_path.exists())

        with OpenMDF(mdf_path) as mdf_file:
            start_datetime = mdf_file.start_time.strftime("%a %b %d %I:%M:%S.%f %p %Y")

            for index, group in enumerate(mdf_file.groups):
                names = [ch.name for ch in group.channels if "time" not in ch.name.lower()]
                if not names:
                    continue

                for channel_name in names:
                    if "." not in channel_name:
                        names.remove(channel_name)
                        break

                if channel_name:
                    channel = mdf_file.get(channel_name, index)
                    if not channel.samples.size:
                        continue

                    with ASCReader(asc_path) as asc_file:
                        for row, msg in enumerate(asc_file):
                            if "Error" in channel_name:
                                self.assertTrue(msg.is_error_frame)
                                continue
                            elif "Remote" in channel_name:
                                self.assertTrue(msg.is_remote_frame)

                            # Evaluate timestamps
                            self.assertAlmostEqual(msg.timestamp, channel.timestamps[row], places=10)

                            # Evaluate other fields
                            for name in names:
                                if name.endswith(".DataBytes"):
                                    self.assertEqual(msg.data, channel[name][row][: msg.dlc].tobytes())
                                else:
                                    data = channel[name].astype("<u4").tolist()
                                    if name.endswith(".ID"):
                                        self.assertEqual(msg.arbitration_id, data[row])
                                    elif name.endswith(".DLC"):
                                        self.assertEqual(msg.dlc, data[row])
                                    elif name.endswith(".Dir"):
                                        self.assertEqual(msg.is_rx, not bool(data[row]))
                                    elif name.endswith(".ESI"):
                                        self.assertEqual(msg.error_state_indicator, bool(data[row]))
                                        self.assertTrue(msg.is_fd)
                                    elif name.endswith(".BRS"):
                                        self.assertEqual(msg.bitrate_switch, bool(data[row]))
                else:
                    continue

            self.assertEqual(asc_file.base, "hex")
            self.assertEqual(asc_file.timestamps_format, "absolute")
            self.assertFalse(asc_file.internal_events_logged)
            self.assertEqual(
                datetime.datetime.strptime(start_datetime, "%a %b %d %I:%M:%S.%f %p %Y"),
                datetime.datetime.strptime(asc_file.date, "%b %d %I:%M:%S.%f %p %Y"),
            )

    def test_output_format_CSV_0(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Events:
            - Ensure that output format is CSV, no checked checkboxes
            - Press PushButton Apply.

        Evaluate:
            - File was created.
            - Ensure that output file has only selected channels
            - Evaluate CSV files
        """
        # Setup
        name = "CSV"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(name, check=False)

        # Expected results
        groups = self.get_selected_groups(channels=self.selected_channels)

        # Event
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        with OpenMDF(self.measurement_file) as mdf_file:
            for index, (group_name, channels_list) in enumerate(groups.items()):
                # prepare expected results
                suffix = group_name.replace(group_name[:4], f".ChannelGroup_{index}").replace(" ", "_") + ".csv"
                csv_file = Path(self.test_workspace, self.default_test_file.replace(".mf4", suffix))
                mdf_channels = mdf_file.select(channels_list)

                # Evaluate if file exist
                self.assertTrue(csv_file.exists(), csv_file)
                # Read file as pandas
                pandas_tab = pd.read_csv(csv_file)
                for channel in mdf_channels:
                    self.assertIn(channel.name, pandas_tab.columns)

                    np.testing.assert_almost_equal(channel.timestamps, pandas_tab.timestamps.values, decimal=9)
                    if np.issubdtype(channel.samples.dtype, np.number):  # problematic conversion
                        np.testing.assert_almost_equal(channel.samples, pandas_tab[channel.name].values, decimal=9)

    def test_output_format_CSV_1(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Events:
            - Ensure that output format is CSV, all checkboxes are checked
            - Press PushButton Apply.

        Evaluate:
            - File was created.
            - Ensure that output file has only selected channels
            - Evaluate CSV files
        """
        # Expected output file name
        csv_path = Path(self.test_workspace, self.default_test_file.replace(".mf4", ".csv"))
        # Expected units
        units = [""]

        # Ensure output format
        name = "CSV"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(name)

        # Event
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(csv_path.exists(), csv_path)

        with OpenMDF(self.measurement_file) as mdf_file:
            mdf_channels = mdf_file.select(self.selected_channels, raw=True)
            units.extend(channel.unit for channel in mdf_channels)

            # Get channels timestamps max, min, difference between extremes
            min_ts = min(channel.timestamps.min() for channel in mdf_channels)
            max_ts = max(channel.timestamps.max() for channel in mdf_channels)
            seconds = int(max_ts - min_ts)
            microseconds = np.floor((max_ts - min_ts - seconds) * pow(10, 6))

            # Read file as pandas
            pandas_tab = pd.read_csv(csv_path, header=[0, 1])
            pandas_tab.timestamps = pd.DatetimeIndex(
                pandas_tab.timestamps.values, yearfirst=True
            ).values - np.datetime64(mdf_file.start_time)

            # Evaluate timestamps min
            self.assertEqual(np.timedelta64(pandas_tab.timestamps.values.min(), "us").item().microseconds, 0)
            self.assertEqual(np.timedelta64(pandas_tab.timestamps.values.min(), "us").item().seconds, 0)
            # Evaluate timestamps max
            self.assertEqual(np.timedelta64(pandas_tab.timestamps.values.max(), "us").item().microseconds, microseconds)
            self.assertEqual(np.timedelta64(pandas_tab.timestamps.values.max(), "us").item().seconds, seconds)

            # Evaluate channels names and units
            for channel, unit in zip(mdf_channels, units, strict=False):
                if channel.display_names:  # DI.<channel name>
                    display_name, _ = zip(*channel.display_names.items(), strict=False)
                    channel_name = display_name[0]
                else:
                    channel_name = channel.name
                self.assertIn(channel_name, pandas_tab.columns.get_level_values(0))
                if channel.unit:
                    self.assertIn(channel.unit, pandas_tab.columns.get_level_values(1))

    def test_output_format_MAT_0(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Events:
            - Ensure that output format is MAT, no checkboxes checked
            - Press PushButton Apply.

        Evaluate:
            - File was created.
            - Ensure that output file has only selected channels
        """
        # Setup
        name = "MAT"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(name, check=False)

        # Expected results
        mat_path = Path(self.test_workspace, self.default_test_file.replace(".mf4", ".mat"))
        groups = self.get_selected_groups(channels=self.selected_channels)

        # Mouse click on Apply button
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(mat_path.exists())

        mat_file = scipy.io.loadmat(str(mat_path))
        to_replace = {" ", ".", "[", "]"}

        with OpenMDF(Path(self.resource, self.default_test_file)) as mdf_file:
            for index, channels_list in enumerate(groups.values()):
                for name in channels_list:
                    signal = mdf_file.get(name)
                    for character in to_replace:
                        name = name.replace(character, "_")
                    name = f"DG{index}_{name}"

                    self.assertIn(name, mat_file)
                    if np.issubdtype(signal.samples.dtype, np.number):
                        self.assertTrue(np.array_equal(signal.samples, mat_file[name][0]))

    def test_output_format_MAT_1(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Events:
            - Ensure that output format is MAT, no checkboxes checked
            - Press PushButton Apply.

        Evaluate:
            - File was created.
            - Ensure that output file has only selected channels
        """
        # Setup
        name = "MAT"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(name)

        # Expected results
        mat_path = Path(self.test_workspace, self.default_test_file.replace(".mf4", ".mat"))

        # Mouse click on Apply button
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(mat_path.exists())

        mat_file = scipy.io.loadmat(str(mat_path))
        to_replace = {" ", ".", "[", "]"}

        with OpenMDF(Path(self.resource, self.default_test_file)) as mdf_file:
            # get common timestamps
            common_timestamps = self.interpolated_timestamps(mdf_file, set(self.selected_channels))

            for channel in self.selected_channels:
                mdf_signal = mdf_file.get(channel, raw=True).interp(
                    common_timestamps,
                    integer_interpolation_mode=self.widget.integer_interpolation,
                    float_interpolation_mode=self.widget.float_interpolation,
                )

                if mdf_signal.display_names:
                    display_name, _ = zip(*mdf_signal.display_names.items(), strict=False)
                    channel = display_name[0]
                for character in to_replace:
                    channel = channel.replace(character, "_")
                if channel + "_0" in mat_file.keys() and "RAT" in mdf_signal.name:
                    channel += "_0"

                self.assertIn(channel, mat_file.keys())
                np.testing.assert_almost_equal(
                    mdf_signal.samples, mat_file[channel][0], decimal=3, err_msg=mdf_signal.name
                )

            common_timestamps -= common_timestamps[0]  # start from 0
            np.testing.assert_almost_equal(mat_file["timestamps"][0], common_timestamps, decimal=3)

    def test_output_format_HDF5_0(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Events:
            - Ensure that output format is HDF5, no checked checkboxes
            - Press PushButton Apply.

        Evaluate:
            - File was created.
            - Ensure that output file has only selected channels
        """
        # Setup
        name = "HDF5"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(name, check=False)

        # Expected results
        hdf5_path = Path(self.test_workspace, self.default_test_file.replace(".mf4", ".hdf"))
        groups = self.get_selected_groups(channels=self.selected_channels)

        # Mouse click on Apply button
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(hdf5_path.exists())

        with OpenHDF5(hdf5_path) as hdf5_file, OpenMDF(self.measurement_file) as mdf_file:
            self.assertEqual(len(hdf5_file.items()) - 1, len(groups))  # 5th item is file path

            for mdf_group, hdf5_group in zip(groups.values(), hdf5_file.values(), strict=False):
                for name in hdf5_group:
                    if name != "time":  # Evaluate channels
                        self.assertIn(name, mdf_group)
                        mdf_channel = mdf_file.select([name])[0]
                        hdf5_channel = hdf5_group.get(name)

                        if np.issubdtype(mdf_channel.samples.dtype, np.number):  # samples are numbers
                            np.testing.assert_almost_equal(mdf_channel.samples, hdf5_channel, decimal=3)
                        else:
                            # Evaluate samples shape
                            self.assertEqual(mdf_channel.samples.size, hdf5_channel.size)

                    else:  # evaluate timestamps
                        hdf5_channel = hdf5_group.get(name)  # for evaluation will be used latest mdf channel from group
                        np.testing.assert_almost_equal(mdf_channel.timestamps, hdf5_channel, decimal=3)

    def test_output_format_HDF5_1(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Events:
            - Ensure that output format is HDF5, all checkboxes are checked
            - Press PushButton Apply.

        Evaluate:
            - File was created.
            - Ensure that output file has only selected channels
        """
        # Expected results
        hdf5_path = Path(self.test_workspace, self.default_test_file.replace(".mf4", ".hdf"))

        # Ensure output format
        name = "HDF5"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(name)

        # Mouse click on Apply button
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(hdf5_path.exists())

        with OpenHDF5(hdf5_path) as hdf5_file, OpenMDF(self.measurement_file) as mdf_file:
            self.assertEqual(len(hdf5_file.items()), 1)  # 1 item

            # Prepare results
            common_timestamps = self.interpolated_timestamps(mdf_file, set(self.selected_channels))
            hdf5_channels = hdf5_file[str(hdf5_path)]

            for channel in self.selected_channels:
                mdf_channel = mdf_file.get(channel, raw=True).interp(
                    common_timestamps,
                    integer_interpolation_mode=self.widget.integer_interpolation,
                    float_interpolation_mode=self.widget.float_interpolation,
                )

                if mdf_channel.display_names:
                    display_name, _ = zip(*mdf_channel.display_names.items(), strict=False)
                    channel = display_name[0]
                if channel + "_0" in hdf5_channels.keys() and "RAT" in mdf_channel.name:
                    channel += "_0"

                hdf5_channel = hdf5_channels.get(channel)

                # Evaluate
                self.assertIn(channel, hdf5_channels)
                np.testing.assert_almost_equal(mdf_channel.samples, hdf5_channel, decimal=3)

    def test_output_format_Parquet_0(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Events:
            - Ensure that output format is Parquet
            - Press PushButton Apply.

        Evaluate:
            - File was created.
            - Ensure that output file has only selected channels with expected values
        """
        # Expected results
        parquet_path = Path(self.test_workspace, self.default_test_file.replace(".mf4", ".parquet"))

        # Setup
        name = "Parquet"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(check=False)

        # Mouse click on Apply button
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate if parquet file exist
        self.assertTrue(parquet_path.exists())

        with OpenMDF(self.measurement_file) as mdf_file:
            pandas_tab = pd.read_parquet(parquet_path)  # open parquet file as pandas tab
            # Get common timestamps
            common_timestamps = self.interpolated_timestamps(mdf_file, set(self.selected_channels))

            # Evaluate
            np.testing.assert_almost_equal(common_timestamps, pandas_tab.index.values, decimal=9)

            for name in self.selected_channels:
                channel = mdf_file.get(name).interp(
                    common_timestamps,
                    integer_interpolation_mode=self.widget.integer_interpolation,
                    float_interpolation_mode=self.widget.float_interpolation,
                )
                if np.issubdtype(channel.samples.dtype, np.number):  # problematic conversion
                    np.testing.assert_almost_equal(channel.samples, pandas_tab[name].values, decimal=9)

    def test_output_format_Parquet_1(self):
        """
        When QThreads are running, event-loops needs to be processed.
        Events:
            - Ensure that output format is Parquet
            - Press PushButton Apply.

        Evaluate:
            - File was created.
            - Ensure that output file has only selected channels with expected values
        """
        # Expected results
        parquet_path = Path(self.test_workspace, self.default_test_file.replace(".mf4", ".parquet"))

        # Setup
        name = "Parquet"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup()

        # Mouse click on Apply button
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate if parquet file exist
        self.assertTrue(parquet_path.exists())

        # Evaluate
        with OpenMDF(self.measurement_file) as mdf_file:
            pandas_tab = pd.read_parquet(parquet_path)  # open parquet file as pandas tab
            # get common timestamp
            common_timestamps = self.interpolated_timestamps(mdf_file, set(self.selected_channels))

            # Evaluate
            np.testing.assert_almost_equal(common_timestamps - common_timestamps[0], pandas_tab.index.values, decimal=9)
            for name in self.selected_channels:
                channel = mdf_file.get(name, raw=True).interp(
                    common_timestamps,
                    integer_interpolation_mode=self.widget.integer_interpolation,
                    float_interpolation_mode=self.widget.float_interpolation,
                )
                if channel.display_names:
                    display_name, _ = zip(*channel.display_names.items(), strict=False)
                    name = display_name[0]
                if name + "_0" in pandas_tab.columns and "RAT" in channel.name:
                    name += "_0"

                if np.issubdtype(channel.samples.dtype, np.number):  # problematic conversion
                    np.testing.assert_almost_equal(channel.samples, pandas_tab[name].values, decimal=3)

    def test_cut_checkbox_0(self):
        """
        Events
            - Check cut group checkbox
            - Set cut start and stop time
            - Check whence checkbox
            - Check cut time from zero checkbox
            - Press Apply bush button

        Evaluate
            - In test workspace folder was created file with same name as original measurement file
            - Start time for mdf file is updated
            - Timestamps start for all channels is equal to 0
            - Timestamps stop for all channels is equal to difference between stop and start time
        """
        # Setup
        name = "MDF"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(check=False)

        # Expected values
        output_file = Path(self.test_workspace, self.default_test_file)

        with OpenMDF(self.measurement_file) as mdf_file:
            self.start_time = mdf_file.start_time.astimezone().replace(tzinfo=None)
            self.ts_min = min(
                [
                    ch.timestamps.min()
                    for ch in mdf_file.iter_channels()
                    if ch.timestamps.min() != 0 and ch.name in self.selected_channels
                ]
            )

        start_cut = randint(1, 42) / 10
        stop_cut = randint(60, 90) / 10
        expected_start_time = self.start_time + datetime.timedelta(milliseconds=(start_cut + self.ts_min) * 1000)

        # Setup
        QtTest.QTest.keyClick(self.widget.cut_group, QtCore.Qt.Key.Key_Space)
        self.widget.cut_start.setValue(start_cut)
        self.widget.cut_stop.setValue(stop_cut)
        for checkbox in self.widget.cut_group.findChildren(QtWidgets.QCheckBox):
            if not checkbox.isChecked():
                self.mouseClick_CheckboxButton(checkbox)
        self.processEvents(0.01)

        # Event
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(output_file.exists())
        with OpenMDF(output_file) as mdf_file:
            time_dif = expected_start_time - mdf_file.start_time
            self.assertEqual(time_dif.microseconds, 0)
            self.assertEqual(time_dif.seconds, 0)
            self.assertEqual(time_dif.days, 0)
            for channel in mdf_file.iter_channels():
                self.assertEqual(channel.timestamps.min(), 0)
                self.assertEqual(channel.timestamps.max(), stop_cut - start_cut)

    def test_cut_checkbox_1(self):
        """
        Events
            - Check cut group checkbox
            - Set cut start and stop time
            - Press Apply bush button

        Evaluate
            - In test workspace folder was created file with same name as original measurement file
            - Start time for mdf file is updated
            - Timestamps start for all channels is equal to 0
            - Timestamps stop for all channels is equal to difference between stop and start time
        """
        # Setup
        name = "MDF"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(check=False)

        # Expected values
        output_file = Path(self.test_workspace, self.default_test_file)

        with OpenMDF(self.measurement_file) as mdf_file:
            self.start_time = mdf_file.start_time
            self.ts_min = min(
                [
                    ch.timestamps.min()
                    for ch in mdf_file.iter_channels()
                    if ch.timestamps.min() != 0 and ch.name in self.selected_channels
                ]
            )

        start_cut = randint(1, 42) / 10
        stop_cut = randint(60, 90) / 10

        # Setup
        QtTest.QTest.keyClick(self.widget.cut_group, QtCore.Qt.Key.Key_Space)
        self.widget.cut_start.setValue(start_cut)
        self.widget.cut_stop.setValue(stop_cut)
        for checkbox in self.widget.cut_group.findChildren(QtWidgets.QCheckBox):
            if checkbox.isChecked():
                self.mouseClick_CheckboxButton(checkbox)
        self.processEvents(0.01)

        # Event
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(output_file.exists())
        with OpenMDF(output_file) as mdf_file:
            self.assertEqual(self.start_time, mdf_file.start_time)

            for channel in mdf_file.iter_channels():
                self.assertEqual(channel.timestamps.min(), start_cut)
                self.assertEqual(channel.timestamps.max(), stop_cut)

    @unittest.skip("FIXME: test keeps failing in CI")
    def test_resample_by_step_0(self):
        """
        Events
            - Check resample group checkbox
            - Set step time
            - Press Apply bush button

        Evaluate
            - In test workspace folder was created file with same name as original measurement file
            - Start time for timestamps is not updated
            - Timestamps step for all channels is the same as step time value
        """
        # Expected values
        output_file = Path(self.test_workspace, self.default_test_file)
        step = randint(1, 100) / 10

        # Setup
        name = "MDF"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(check=False)

        QtTest.QTest.keyClick(self.widget.resample_group, QtCore.Qt.Key.Key_Space)
        self.widget.raster_type_step.setChecked(True)
        self.widget.raster.setValue(step)
        for checkbox in self.widget.resample_group.findChildren(QtWidgets.QCheckBox):
            if checkbox.isChecked():
                self.mouseClick_CheckboxButton(checkbox)
        self.processEvents(0.01)

        # Event
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(output_file.exists())
        with OpenMDF(output_file) as mdf_file:
            for channel in mdf_file.iter_channels():
                size = ceil((channel.timestamps.max() - channel.timestamps.min()) / step) + 1  # + min
                self.assertEqual(size, channel.timestamps.size)
                for i in range(size):
                    self.assertAlmostEqual(channel.timestamps[i], channel.timestamps.min() + step * i, places=12)

    @unittest.skip("FIXME: test keeps failing in CI")
    def test_resample_by_step_1(self):
        """
        Events
            - Check resample group checkbox
            - Set step time
            - Check Time from zero checkbox
            - Press Apply bush button

        Evaluate
            - In test workspace folder was created file with same name as original measurement file
            - Start time for timestamps is 0
            - Timestamps step for all channels is the same as step time value
        """
        # Expected values
        output_file = Path(self.test_workspace, self.default_test_file)
        step = randint(1, 100) / 10

        # Setup
        name = "MDF"
        self.widget.output_format.setCurrentText(name)
        self.generic_setup(check=False)

        QtTest.QTest.keyClick(self.widget.resample_group, QtCore.Qt.Key.Key_Space)
        self.widget.raster_type_step.setChecked(True)
        self.widget.raster.setValue(step)
        for checkbox in self.widget.resample_group.findChildren(QtWidgets.QCheckBox):
            if not checkbox.isChecked():
                self.mouseClick_CheckboxButton(checkbox)
        self.processEvents(0.01)

        # Event
        self.mouse_click_on_btn_with_progress(self.tested_btn)

        # Evaluate
        self.assertTrue(output_file.exists())
        with OpenMDF(output_file) as mdf_file:
            for channel in mdf_file.iter_channels():
                size = ceil((channel.timestamps.max() - channel.timestamps.min()) / step) + 1  # + min
                self.assertEqual(size, channel.timestamps.size)
                for i in range(size):
                    self.assertAlmostEqual(channel.timestamps[i], step * i, places=12)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'main_window.ui'
##
## Created by: Qt User Interface Compiler version 6.7.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QMainWindow, QMenuBar,
    QSizePolicy, QStackedWidget, QTabWidget, QVBoxLayout,
    QWidget)
from . import resource_rc

class Ui_PyMDFMainWindow(object):
    def setupUi(self, PyMDFMainWindow):
        if not PyMDFMainWindow.objectName():
            PyMDFMainWindow.setObjectName(u"PyMDFMainWindow")
        PyMDFMainWindow.resize(800, 723)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(PyMDFMainWindow.sizePolicy().hasHeightForWidth())
        PyMDFMainWindow.setSizePolicy(sizePolicy)
        icon = QIcon()
        icon.addFile(u":/asammdf.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        PyMDFMainWindow.setWindowIcon(icon)
        self.action_memory_minimum = QAction(PyMDFMainWindow)
        self.action_memory_minimum.setObjectName(u"action_memory_minimum")
        self.action_memory_minimum.setCheckable(True)
        self.action_memory_full = QAction(PyMDFMainWindow)
        self.action_memory_full.setObjectName(u"action_memory_full")
        self.action_memory_full.setCheckable(True)
        self.action_memory_low = QAction(PyMDFMainWindow)
        self.action_memory_low.setObjectName(u"action_memory_low")
        self.action_memory_low.setCheckable(True)
        self.centralwidget = QWidget(PyMDFMainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setSpacing(1)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(1, 1, 1, 1)
        self.stackedWidget = QStackedWidget(self.centralwidget)
        self.stackedWidget.setObjectName(u"stackedWidget")
        self.stackedWidget.setFrameShape(QFrame.NoFrame)
        self.stackedWidgetPage1 = QWidget()
        self.stackedWidgetPage1.setObjectName(u"stackedWidgetPage1")
        self.verticalLayout_2 = QVBoxLayout(self.stackedWidgetPage1)
        self.verticalLayout_2.setSpacing(1)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(1, 1, 1, 1)
        self.files = QTabWidget(self.stackedWidgetPage1)
        self.files.setObjectName(u"files")
        self.files.setDocumentMode(False)
        self.files.setTabsClosable(True)

        self.verticalLayout_2.addWidget(self.files)

        self.stackedWidget.addWidget(self.stackedWidgetPage1)

        self.verticalLayout.addWidget(self.stackedWidget)

        PyMDFMainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(PyMDFMainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 800, 21))
        PyMDFMainWindow.setMenuBar(self.menubar)

        self.retranslateUi(PyMDFMainWindow)

        self.stackedWidget.setCurrentIndex(0)
        self.files.setCurrentIndex(-1)


        QMetaObject.connectSlotsByName(PyMDFMainWindow)
    # setupUi

    def retranslateUi(self, PyMDFMainWindow):
        PyMDFMainWindow.setWindowTitle(QCoreApplication.translate("PyMDFMainWindow", u"asammdf", None))
        self.action_memory_minimum.setText(QCoreApplication.translate("PyMDFMainWindow", u"minimum", None))
#if QT_CONFIG(tooltip)
        self.action_memory_minimum.setToolTip(QCoreApplication.translate("PyMDFMainWindow", u"Minimal memory usage by loading only the nedded block addresses", None))
#endif // QT_CONFIG(tooltip)
        self.action_memory_full.setText(QCoreApplication.translate("PyMDFMainWindow", u"full", None))
#if QT_CONFIG(tooltip)
        self.action_memory_full.setToolTip(QCoreApplication.translate("PyMDFMainWindow", u"Load all blocks in the RAM", None))
#endif // QT_CONFIG(tooltip)
        self.action_memory_low.setText(QCoreApplication.translate("PyMDFMainWindow", u"low", None))
#if QT_CONFIG(tooltip)
        self.action_memory_low.setToolTip(QCoreApplication.translate("PyMDFMainWindow", u"Load metdata block in RAM but leave the samples on disk", None))
#endif // QT_CONFIG(tooltip)
    # retranslateUi



from bz2 import BZ2File
from collections.abc import Sequence
from gzip import GzipFile
from io import BufferedRandom, BufferedReader, BufferedWriter, BytesIO
from mmap import mmap
from os import PathLike
from typing import Any, Literal, TYPE_CHECKING, Union
from zipfile import ZipFile

from canmatrix import CanMatrix
from numpy.typing import NDArray

if TYPE_CHECKING:
    from .blocks import v2_v3_blocks, v4_blocks
    from .blocks.mdf_v2 import MDF2
    from .blocks.mdf_v3 import MDF3
    from .blocks.mdf_v4 import MDF4
    from .blocks.source_utils import Source


StrPathType = Union[str, "PathLike[str]"]
StrOrBytesPathType = Union[str, bytes, "PathLike[str]", "PathLike[bytes]"]
ReadableBufferType = Union[BufferedRandom, BufferedReader, BytesIO, mmap]
WritableBufferType = Union[BufferedRandom, BufferedWriter, BytesIO, mmap]

# asammdf specific types

BusType = Literal["CAN", "LIN"]
ChannelConversionType = Union["v2_v3_blocks.ChannelConversion", "v4_blocks.ChannelConversion"]
ChannelGroupType = Union["v2_v3_blocks.ChannelGroup", "v4_blocks.ChannelGroup"]
ChannelsType = Union[Sequence[str], Sequence[tuple[str | None, int, int]], Sequence[tuple[str, int]]]
ChannelType = Union["v2_v3_blocks.Channel", "v4_blocks.Channel"]
CompressionType = Union[Literal[0, 1, 2, 3, 4, 5], "v4_blocks.CompressionAlgorithm"]
DataGroupType = Union["v2_v3_blocks.DataGroup", "v4_blocks.DataGroup"]
DbcFileType = tuple[StrPathType | CanMatrix, int]
EmptyChannelsType = Literal["skip", "zeros"]
FloatInterpolationModeType = Literal[0, 1]
InputType = Union[BufferedReader, BytesIO, StrPathType, ZipFile, BZ2File, GzipFile]
IntInterpolationModeType = Literal[0, 1, 2]
MDF_v2_v3_v4 = Union["MDF2", "MDF3", "MDF4"]
RasterType = Union[float, str, NDArray[Any]]
SourceType = Union["v2_v3_blocks.ChannelExtension", "v4_blocks.SourceInformation", "Source"]
SyncType = Literal[0, 1, 2, 3, 4]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'multi_search_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.7.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDialog, QGridLayout,
    QHBoxLayout, QLabel, QLineEdit, QListWidget,
    QListWidgetItem, QPushButton, QSizePolicy, QSpacerItem,
    QWidget)

from asammdf.gui.widgets.list import MinimalListWidget
from . import resource_rc

class Ui_MultiSearchDialog(object):
    def setupUi(self, MultiSearchDialog):
        if not MultiSearchDialog.objectName():
            MultiSearchDialog.setObjectName(u"MultiSearchDialog")
        MultiSearchDialog.resize(1028, 549)
        MultiSearchDialog.setSizeGripEnabled(True)
        self.grid_layout = QGridLayout(MultiSearchDialog)
        self.grid_layout.setSpacing(1)
        self.grid_layout.setObjectName(u"grid_layout")
        self.grid_layout.setContentsMargins(1, 1, 1, 1)
        self.match_kind = QComboBox(MultiSearchDialog)
        self.match_kind.addItem("")
        self.match_kind.addItem("")
        self.match_kind.setObjectName(u"match_kind")

        self.grid_layout.addWidget(self.match_kind, 0, 0, 1, 1)

        self.status = QLabel(MultiSearchDialog)
        self.status.setObjectName(u"status")

        self.grid_layout.addWidget(self.status, 10, 0, 1, 1)

        self.matches = QListWidget(MultiSearchDialog)
        self.matches.setObjectName(u"matches")

        self.grid_layout.addWidget(self.matches, 6, 0, 2, 1)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.grid_layout.addItem(self.verticalSpacer_2, 7, 1, 1, 1)

        self.selection = MinimalListWidget(MultiSearchDialog)
        self.selection.setObjectName(u"selection")
        self.selection.setMinimumSize(QSize(500, 0))

        self.grid_layout.addWidget(self.selection, 6, 2, 3, 4)

        self.label = QLabel(MultiSearchDialog)
        self.label.setObjectName(u"label")

        self.grid_layout.addWidget(self.label, 2, 2, 1, 1)

        self.search_box = QLineEdit(MultiSearchDialog)
        self.search_box.setObjectName(u"search_box")

        self.grid_layout.addWidget(self.search_box, 2, 0, 1, 1)

        self.show_measurement_list_btn = QPushButton(MultiSearchDialog)
        self.show_measurement_list_btn.setObjectName(u"show_measurement_list_btn")
        icon = QIcon()
        icon.addFile(u":/info.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.show_measurement_list_btn.setIcon(icon)

        self.grid_layout.addWidget(self.show_measurement_list_btn, 0, 2, 1, 2)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(1)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.cancel_btn = QPushButton(MultiSearchDialog)
        self.cancel_btn.setObjectName(u"cancel_btn")

        self.horizontalLayout.addWidget(self.cancel_btn)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self.apply_btn = QPushButton(MultiSearchDialog)
        self.apply_btn.setObjectName(u"apply_btn")

        self.horizontalLayout.addWidget(self.apply_btn)

        self.horizontalLayout.setStretch(1, 1)

        self.grid_layout.addLayout(self.horizontalLayout, 9, 0, 1, 6)

        self.add_btn = QPushButton(MultiSearchDialog)
        self.add_btn.setObjectName(u"add_btn")
        self.add_btn.setFocusPolicy(Qt.FocusPolicy.TabFocus)
        icon1 = QIcon()
        icon1.addFile(u":/left.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.add_btn.setIcon(icon1)

        self.grid_layout.addWidget(self.add_btn, 6, 1, 1, 1)

        self.case_sensitivity = QComboBox(MultiSearchDialog)
        self.case_sensitivity.addItem("")
        self.case_sensitivity.addItem("")
        self.case_sensitivity.setObjectName(u"case_sensitivity")

        self.grid_layout.addWidget(self.case_sensitivity, 1, 0, 1, 1)

        self.grid_layout.setColumnStretch(0, 1)
        QWidget.setTabOrder(self.match_kind, self.case_sensitivity)
        QWidget.setTabOrder(self.case_sensitivity, self.search_box)
        QWidget.setTabOrder(self.search_box, self.matches)
        QWidget.setTabOrder(self.matches, self.add_btn)
        QWidget.setTabOrder(self.add_btn, self.selection)
        QWidget.setTabOrder(self.selection, self.apply_btn)
        QWidget.setTabOrder(self.apply_btn, self.cancel_btn)
        QWidget.setTabOrder(self.cancel_btn, self.show_measurement_list_btn)

        self.retranslateUi(MultiSearchDialog)

        self.match_kind.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(MultiSearchDialog)
    # setupUi

    def retranslateUi(self, MultiSearchDialog):
        MultiSearchDialog.setWindowTitle(QCoreApplication.translate("MultiSearchDialog", u"Dialog", None))
        self.match_kind.setItemText(0, QCoreApplication.translate("MultiSearchDialog", u"Wildcard", None))
        self.match_kind.setItemText(1, QCoreApplication.translate("MultiSearchDialog", u"Regex", None))

        self.status.setText("")
        self.label.setText(QCoreApplication.translate("MultiSearchDialog", u"Final selection", None))
        self.search_box.setText("")
        self.show_measurement_list_btn.setText(QCoreApplication.translate("MultiSearchDialog", u"Show measurement list", None))
        self.cancel_btn.setText(QCoreApplication.translate("MultiSearchDialog", u"Cancel", None))
        self.apply_btn.setText(QCoreApplication.translate("MultiSearchDialog", u"Apply", None))
        self.add_btn.setText("")
        self.case_sensitivity.setItemText(0, QCoreApplication.translate("MultiSearchDialog", u"Case insensitive", None))
        self.case_sensitivity.setItemText(1, QCoreApplication.translate("MultiSearchDialog", u"Case sensitive", None))

    # retranslateUi



import copy
import inspect
import json
import pathlib
from unittest import mock

from PySide6 import QtCore, QtTest

from asammdf.gui.dialogs.functions_manager import FunctionsManagerDialog
from test.asammdf.gui.resources.functions import (
    Function2,
    gray2dec,
    maximum,
    rpm_to_rad_per_second,
    UnresolvedVariable,
    WrongDefinition,
)
from test.asammdf.gui.test_base import TestBase


class TestPushButtons(TestBase):
    def setUp(self) -> None:
        super().setUp()
        self.fm = FunctionsManagerDialog({})

    def test_PushButton_Add(self):
        """
        Test Scope:
            - Ensure that Functions can be added.
        Precondition:
            - Open Function Manager Widget
        Events:
            - Press PushButton "Add"
            - Press PushButton "Add"
        Evaluate:
            - Ensure that two items will be added in functions list
        """
        # Events:
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        self.assertEqual(2, self.fm.widget.functions_list.count())
        for index in range(3, 1):
            self.assertEqual(f"Function{index}", self.fm.widget.functions_list.item(index - 1))

    def test_PushButton_DeleteAll(self):
        """
        Test Scope:
            - Ensure that all Functions are deleted via PushButton DeleteAll.
        Preconditions:
            - Open Function Manager Widget
        Events:
            - Press PushButton "Add"
            - Press PushButton "Add"
            - Press PushButton "DeleteAll"
        Evaluate:
            - Evaluate that items are deleted from functions list.
        """
        # Events:
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)

        self.assertEqual(2, self.fm.widget.functions_list.count())
        for index in range(3, 1):
            self.assertEqual(f"Function{index}", self.fm.widget.functions_list.item(index - 1))

        QtTest.QTest.mouseClick(self.fm.widget.erase_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        self.assertEqual(0, self.fm.widget.functions_list.count())

    def test_PushButton_LoadDefinitions_0(self):
        """
        Test Scope:
            - Ensure that Functions from '.def' file are loaded in function list
        Preconditions:
            - Open Function Manager Widget
        Events:
            - Press PushButton "Load Definitions"
                - Simulate that valid path is provided.
        Evaluate:
            - Evaluate that files are loaded into function list.
        """
        definition_file = pathlib.Path(self.resource, "functions_definitions.def")
        with mock.patch(
            "asammdf.gui.widgets.functions_manager.QtWidgets.QFileDialog.getOpenFileName"
        ) as mc_getOpenFileName:
            mc_getOpenFileName.return_value = definition_file, None
            QtTest.QTest.mouseClick(self.fm.widget.import_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        self.assertEqual(4, self.fm.widget.functions_list.count())
        self.assertEqual("Function1", self.fm.widget.functions_list.item(0).text())
        self.assertEqual("gray2dec", self.fm.widget.functions_list.item(1).text())
        self.assertEqual("maximum", self.fm.widget.functions_list.item(2).text())
        self.assertEqual("rpm_to_rad_per_second", self.fm.widget.functions_list.item(3).text())

    def test_PushButton_LoadDefinitions_1(self):
        """
        Test Scope:
            - Ensure that Functions from nonexistent '.def' file is handled.
        Preconditions:
            - Open Function Manager Widget
        Events:
            - Press PushButton "Load Definitions"
                - Simulate that valid path is provided but file does not exist.
        Evaluate:
            - Evaluate that functions from file are loaded into function list.
        """
        definition_file = pathlib.Path(self.test_workspace, "nonexistent.def")
        with mock.patch(
            "asammdf.gui.widgets.functions_manager.QtWidgets.QFileDialog.getOpenFileName"
        ) as mc_getOpenFileName:
            mc_getOpenFileName.return_value = definition_file, None
            QtTest.QTest.mouseClick(self.fm.widget.import_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        self.assertEqual(0, self.fm.widget.functions_list.count())

    def test_PushButton_LoadDefinitions_2(self):
        """
        Test Scope:
            - Ensure that case where import dialog is closed, is handled.
        Preconditions:
            - Open Function Manager Widget
        Events:
            - Press PushButton "Load Definitions"
                - Simulate that no valid path is provided.
        Evaluate:
            - Evaluate that no functions is loaded to function list.
        """
        with mock.patch(
            "asammdf.gui.widgets.functions_manager.QtWidgets.QFileDialog.getOpenFileName"
        ) as mc_getOpenFileName:
            mc_getOpenFileName.return_value = None, None
            QtTest.QTest.mouseClick(self.fm.widget.import_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        self.assertEqual(0, self.fm.widget.functions_list.count())

    def test_PushButton_LoadDefinitions_3(self):
        """
        Test Scope:
            - Ensure that Functions from '.def' file are loaded in function list
        Preconditions:
            - Open Function Manager Widget
        Events:
            - Press PushButton "Load Definitions"
                - Simulate that valid path is provided but file does not have correct structure.
        Evaluate:
            - Evaluate that no functions is loaded to function list.
        """
        definition_file = pathlib.Path(self.test_workspace, "invalid.def")
        with open(definition_file, "w+") as fpw:
            fpw.write(self.id())

        with mock.patch(
            "asammdf.gui.widgets.functions_manager.QtWidgets.QFileDialog.getOpenFileName"
        ) as mc_getOpenFileName:
            mc_getOpenFileName.return_value = definition_file, None
            QtTest.QTest.mouseClick(self.fm.widget.import_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        self.assertEqual(0, self.fm.widget.functions_list.count())
        self.mc_ErrorDialog.assert_called()

    def test_PushButton_LoadDefinitions_4(self):
        """
        Test Scope:
            - Ensure that Functions from file with extension different from '.def' are not loaded in function list
        Preconditions:
            - Open Function Manager Widget
        Events:
            - Press PushButton "Load Definitions"
                - Simulate that valid path is provided but file does not have correct structure.
        Evaluate:
            - Evaluate that no functions is loaded to function list.
        """
        definition_file = pathlib.Path(self.test_workspace, "wrong_extension.deff")
        with open(definition_file, "w+") as fpw:
            json.dump(
                {"Function1": "def Function1(t=0):\n    return 0"},
                fpw,
                sort_keys=True,
                indent=2,
            )

        with mock.patch(
            "asammdf.gui.widgets.functions_manager.QtWidgets.QFileDialog.getOpenFileName"
        ) as mc_getOpenFileName:
            mc_getOpenFileName.return_value = definition_file, None
            QtTest.QTest.mouseClick(self.fm.widget.import_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        self.assertEqual(0, self.fm.widget.functions_list.count())

    def test_PushButton_SaveDefinitions(self):
        """
        Test Scope:
            - Ensure that Functions from '.def' file are loaded in function list
        Preconditions:
            - Open Function Manager Widget
        Events:
            - Press PushButton Add
            - Press PushButton Save Definitions
        Evaluate:
            - Evaluate that dialog is open and file can be saved.
            - Evaluate that new function is present in saved file.
        """
        # Event
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)
        saved_file = pathlib.Path(self.test_workspace, f"{self.id()}.def")
        with mock.patch(
            "asammdf.gui.widgets.functions_manager.QtWidgets.QFileDialog.getSaveFileName"
        ) as mc_getSaveFileName:
            mc_getSaveFileName.return_value = str(saved_file), None
            QtTest.QTest.mouseClick(self.fm.widget.export_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        self.assertTrue(saved_file.exists())
        with open(saved_file) as fpr:
            content = json.load(fpr)
            self.assertDictEqual(
                content, {"Function1": "def Function1(t=0):\n    return 0", "__global_variables__": ""}
            )

    def test_PushButton_CheckSyntax_0(self):
        """
        Test Scope:
            - Ensure that valid Python Syntax is detected.
        Events:
            - Press PushButton Add
            - Edit Function1 content. Add valid python syntax
            - Press PushButton Check Syntax
        Evaluate:
            - Evaluate that python function is evaluated as valid.
        """
        # Event
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)

        for f in (
            gray2dec,
            maximum,
            rpm_to_rad_per_second,
        ):
            with self.subTest(f"{self.id}_{f.__name__}"):
                self.mc_ErrorDialog.reset_mock()

                source = inspect.getsource(f)
                self.fm.widget.function_definition.clear()
                self.fm.widget.function_definition.setPlainText(source)

                with mock.patch("asammdf.gui.utils.MessageBox.information") as mc_information:
                    QtTest.QTest.mouseClick(
                        self.fm.widget.check_syntax_btn,
                        QtCore.Qt.MouseButton.LeftButton,
                    )

                    # Evaluate
                    self.mc_ErrorDialog.assert_not_called()
                    mc_information.assert_called()

    def test_PushButton_CheckSyntax_1(self):
        """
        Test Scope:
            - Ensure that invalid Python Syntax is detected.
        Events:
            - Press PushButton Add
            - Edit Function1 content. Add valid python syntax
            - Press PushButton Check Syntax
        Evaluate:
            - Evaluate that python function is evaluated as valid.
        """
        # Event
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)

        for f in (
            Function2,
            UnresolvedVariable,
            WrongDefinition,
        ):
            with self.subTest(f"{self.id}_{f.__name__}"):
                self.mc_ErrorDialog.reset_mock()

                source = inspect.getsource(f)
                self.fm.widget.function_definition.clear()
                self.fm.widget.function_definition.setPlainText(source)

                with mock.patch("asammdf.gui.utils.MessageBox.information") as mc_information:
                    QtTest.QTest.mouseClick(
                        self.fm.widget.check_syntax_btn,
                        QtCore.Qt.MouseButton.LeftButton,
                    )

                    # Evaluate
                    mc_information.assert_not_called()
                    self.mc_ErrorDialog.assert_called()

    def test_PushButton_StoreFunctionChanges_0(self):
        """
        Test Scope:
            - Ensure that function changes are saved.
        Events:
            - Press PushButton Add
            - Edit Function content
            - Press PushButton Store Function Changes
            - Press PushButton Save Definitions
        Evaluate:
            - Evaluate that name of the Function was changed.
            - Evaluate that content of the function was changed.
        """
        # Events:
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)
        self.mouseClick_WidgetItem(self.fm.widget.functions_list.item(0))

        source = inspect.getsource(maximum)
        self.fm.widget.function_definition.clear()
        self.fm.widget.function_definition.setPlainText(source)

        QtTest.QTest.mouseClick(self.fm.widget.store_btn, QtCore.Qt.MouseButton.LeftButton)

        saved_file = pathlib.Path(self.test_workspace, f"{self.id()}.def")
        with mock.patch(
            "asammdf.gui.widgets.functions_manager.QtWidgets.QFileDialog.getSaveFileName"
        ) as mc_getSaveFileName:
            mc_getSaveFileName.return_value = str(saved_file), None
            QtTest.QTest.mouseClick(self.fm.widget.export_btn, QtCore.Qt.MouseButton.LeftButton)

        # Evaluate
        self.assertTrue(saved_file.exists())
        with open(saved_file) as fpr:
            content = json.load(fpr)
            self.assertIn(maximum.__name__, content)
            self.assertIn(content["maximum"], source)

    def test_PushButton_StoreFunctionChanges_1(self):
        """
        Test Scope:
            - Ensure that function changes are saved.
        Events:
            - Press PushButton Add
            - Press PushButton Add
            - Edit 2nd Function and rename as 1st Function
            - Press PushButton Store Function Changes
        Evaluate:
            - Evaluate that overwriting is not possible
        """
        # Events:
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)

        self.mouseClick_WidgetItem(self.fm.widget.functions_list.item(0))
        function1 = self.fm.widget.function_definition.toPlainText()

        self.mouseClick_WidgetItem(self.fm.widget.functions_list.item(1))
        self.fm.widget.function_definition.clear()
        self.fm.widget.function_definition.setPlainText(function1)

        with mock.patch("asammdf.gui.widgets.functions_manager.MessageBox.information") as mc_information:
            QtTest.QTest.mouseClick(self.fm.widget.store_btn, QtCore.Qt.MouseButton.LeftButton)

            mc_information.assert_called()

    def test_PushButton_Apply(self):
        """
        Test Scope:
            - Ensure that definitions are saved when apply button is pressed.
        Events:
            - Press PushButton Add
            - Press PushButton Add
            - Update content of current function
            - Press PushButton Store
            - Press PushButton 'Apply'
        Evaluate:
            - Evaluate that definitions attribute is the same.
        """
        # Events
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)
        definitions = copy.deepcopy(self.fm.widget.definitions)

        self.fm.widget.function_definition.clear()
        source = inspect.getsource(maximum)
        self.fm.widget.function_definition.setPlainText(source)
        QtTest.QTest.mouseClick(self.fm.widget.store_btn, QtCore.Qt.MouseButton.LeftButton)
        QtTest.QTest.mouseClick(self.fm.apply_btn, QtCore.Qt.MouseButton.LeftButton)

        self.assertEqual("apply", self.fm.pressed_button)
        self.assertEqual(2, len(self.fm.modified_definitions))

    def test_PushButton_Cancel(self):
        """
        Test Scope:
            - Ensure that definitions are not saved when cancel button is pressed.
        Events:
            - Press PushButton Add
            - Press PushButton Add
            - Update content of current function
            - Press PushButton Store
            - Press PushButton Cancel
        Evaluate:
            - Evaluate that definitions attribute is the same.
        """
        # Events
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)

        self.fm.widget.function_definition.clear()
        source = inspect.getsource(maximum)
        self.fm.widget.function_definition.setPlainText(source)
        QtTest.QTest.mouseClick(self.fm.widget.store_btn, QtCore.Qt.MouseButton.LeftButton)
        QtTest.QTest.mouseClick(self.fm.cancel_btn, QtCore.Qt.MouseButton.LeftButton)

        self.assertEqual("cancel", self.fm.pressed_button)
        self.assertDictEqual({}, self.fm.modified_definitions)


class TestTreeWidget(TestBase):
    def setUp(self) -> None:
        super().setUp()
        self.fm = FunctionsManagerDialog({"Function10": "def Function10(t=0):\n    return 0"})

    def test_KeyPress_Delete(self):
        """
        Test Scope:
            - Ensure that functions can be deleted from function list.
        Events:
            - Press PushButton Add
            - Press PushButton Add
            - Press PushButton Add
            - Select "Function2"
            - Press Key Delete
        Evaluate:
            - Evaluate that "Function2" is no longer part of function list.
        """
        # Events:
        for _ in range(3):
            QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)
        self.assertEqual(4, self.fm.widget.functions_list.count())

        self.mouseClick_WidgetItem(self.fm.widget.functions_list.item(2))
        QtTest.QTest.keyClick(self.fm.widget.functions_list, QtCore.Qt.Key.Key_Delete)

        # Evaluate
        self.assertEqual(3, self.fm.widget.functions_list.count())
        self.assertNotEqual("Function2", self.fm.widget.functions_list.item(0))
        self.assertNotEqual("Function2", self.fm.widget.functions_list.item(1))

    def test_FunctionDefinition_ContentUpdate(self):
        """
        Test Scope:
            - Ensure that content is updated in Function Definition when Function is changed in FunctionList
        Events:
            - Press PushButton Add
            - Select Function1
            - Select Function10
        Evaluate:
            - Evaluate that content of function definition is changed.
        """
        # Event
        QtTest.QTest.mouseClick(self.fm.widget.add_btn, QtCore.Qt.MouseButton.LeftButton)
        self.assertEqual(2, self.fm.widget.functions_list.count())

        self.mouseClick_WidgetItem(self.fm.widget.functions_list.item(0))
        function1 = self.fm.widget.function_definition.toPlainText()

        self.mouseClick_WidgetItem(self.fm.widget.functions_list.item(1))
        function2 = self.fm.widget.function_definition.toPlainText()

        # Evaluate
        self.assertNotEqual(function1, function2)



from pathlib import Path

import numpy as np

from asammdf import MDF, Signal, SUPPORTED_VERSIONS
import asammdf.blocks.v2_v3_blocks as v3b
import asammdf.blocks.v2_v3_constants as v3c
import asammdf.blocks.v4_blocks as v4b
import asammdf.blocks.v4_constants as v4c

SUPPORTED_VERSIONS = SUPPORTED_VERSIONS[1:]

cycles = 500
channels_count = 20
array_channels_count = 20


def get_test_data(filename=""):
    """
    Utility functions needed by all test scripts.
    """
    return Path(__file__).resolve().parent.joinpath("/data/", filename)


def generate_test_file(tmpdir, version="4.10"):
    mdf = MDF(version=version)

    if version <= "3.30":
        filename = Path(tmpdir) / f"big_test_{version}.mdf"
    else:
        filename = Path(tmpdir) / f"big_test_{version}.mf4"

    if filename.exists():
        return filename

    t = np.arange(cycles, dtype=np.float64)

    cls = v4b.ChannelConversion if version >= "4.00" else v3b.ChannelConversion

    # no conversion
    sigs = []
    for i in range(channels_count):
        sig = Signal(
            np.ones(cycles, dtype=np.uint64) * i,
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=None,
            comment=f"Unsigned int 16bit channel {i}",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # linear
    sigs = []
    for i in range(channels_count):
        conversion = {
            "conversion_type": v4c.CONVERSION_TYPE_LIN if version >= "4.00" else v3c.CONVERSION_TYPE_LINEAR,
            "a": float(i),
            "b": -0.5,
        }
        sig = Signal(
            np.ones(cycles, dtype=np.int64),
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=cls(**conversion),
            comment=f"Signed 16bit channel {i} with linear conversion",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # algebraic
    sigs = []
    for i in range(channels_count):
        conversion = {
            "conversion_type": v4c.CONVERSION_TYPE_ALG if version >= "4.00" else v3c.CONVERSION_TYPE_FORMULA,
            "formula": f"{i} * sin(X)",
        }
        sig = Signal(
            np.arange(cycles, dtype=np.int32) / 100.0,
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=cls(**conversion),
            comment=f"Sinus channel {i} with algebraic conversion",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # rational
    sigs = []
    for i in range(channels_count):
        conversion = {
            "conversion_type": v4c.CONVERSION_TYPE_RAT if version >= "4.00" else v3c.CONVERSION_TYPE_RAT,
            "P1": 0,
            "P2": i,
            "P3": -0.5,
            "P4": 0,
            "P5": 0,
            "P6": 1,
        }
        sig = Signal(
            np.ones(cycles, dtype=np.int64),
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=cls(**conversion),
            comment=f"Channel {i} with rational conversion",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # string
    sigs = []
    encoding = "latin-1" if version < "4.00" else "utf-8"
    for i in range(channels_count):
        sig = [f"Channel {i} sample {j}".encode(encoding) for j in range(cycles)]
        sig = Signal(
            np.array(sig),
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            comment=f"String channel {i}",
            raw=True,
            encoding=encoding,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # byte array
    sigs = []
    ones = np.ones(cycles, dtype=np.dtype("(8,)u1"))
    for i in range(channels_count):
        sig = Signal(
            ones * i,
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            comment=f"Byte array channel {i}",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # value to text
    sigs = []
    ones = np.ones(cycles, dtype=np.uint64)
    conversion = {
        "raw": np.arange(255, dtype=np.float64),
        "phys": np.array([f"Value {i}".encode("ascii") for i in range(255)]),
        "conversion_type": v4c.CONVERSION_TYPE_TABX if version >= "4.00" else v3c.CONVERSION_TYPE_TABX,
        "links_nr": 260,
        "ref_param_nr": 255,
    }

    for i in range(255):
        conversion[f"val_{i}"] = conversion[f"param_val_{i}"] = conversion["raw"][i]
        conversion[f"text_{i}"] = conversion["phys"][i]
    conversion[f"text_{255}"] = "Default"

    for i in range(channels_count):
        sig = Signal(
            ones * i,
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            comment=f"Value to text channel {i}",
            conversion=cls(**conversion),
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    name = mdf.save(filename, overwrite=True)
    mdf.close()


def generate_arrays_test_file(tmpdir):
    version = "4.10"
    mdf = MDF(version=version)
    filename = Path(tmpdir) / f"arrays_test_{version}.mf4"

    if filename.exists():
        return filename

    t = np.arange(cycles, dtype=np.float64)

    # lookup tabel with axis
    sigs = []
    for i in range(array_channels_count):
        samples = [
            np.ones((cycles, 2, 3), dtype=np.uint64) * i,
            np.ones((cycles, 2), dtype=np.uint64) * i,
            np.ones((cycles, 3), dtype=np.uint64) * i,
        ]

        types = [
            (f"Channel_{i}", "(2, 3)<u8"),
            (f"channel_{i}_axis_1", "(2, )<u8"),
            (f"channel_{i}_axis_2", "(3, )<u8"),
        ]

        sig = Signal(
            np.rec.fromarrays(samples, dtype=np.dtype(types)),
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=None,
            comment=f"Array channel {i}",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # lookup tabel with default axis
    sigs = []
    for i in range(array_channels_count):
        samples = [np.ones((cycles, 2, 3), dtype=np.uint64) * i]

        types = [(f"Channel_{i}", "(2, 3)<u8")]

        sig = Signal(
            np.rec.fromarrays(samples, dtype=np.dtype(types)),
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=None,
            comment=f"Array channel {i} with default axis",
            raw=True,
        )
        sigs.append(sig)
    mdf.append(sigs, common_timebase=True)

    # structure channel composition
    sigs = []
    for i in range(array_channels_count):
        samples = [
            np.ones(cycles, dtype=np.uint8) * i,
            np.ones(cycles, dtype=np.uint16) * i,
            np.ones(cycles, dtype=np.uint32) * i,
            np.ones(cycles, dtype=np.uint64) * i,
            np.ones(cycles, dtype=np.int8) * i,
            np.ones(cycles, dtype=np.int16) * i,
            np.ones(cycles, dtype=np.int32) * i,
            np.ones(cycles, dtype=np.int64) * i,
        ]

        types = [
            (f"struct_{i}_channel_0", np.uint8),
            (f"struct_{i}_channel_1", np.uint16),
            (f"struct_{i}_channel_2", np.uint32),
            (f"struct_{i}_channel_3", np.uint64),
            (f"struct_{i}_channel_4", np.int8),
            (f"struct_{i}_channel_5", np.int16),
            (f"struct_{i}_channel_6", np.int32),
            (f"struct_{i}_channel_7", np.int64),
        ]

        sig = Signal(
            np.rec.fromarrays(samples, dtype=np.dtype(types)),
            t,
            name=f"Channel_{i}",
            unit=f"unit_{i}",
            conversion=None,
            comment=f"Structure channel composition {i}",
            raw=True,
        )
        sigs.append(sig)

    mdf.append(sigs, common_timebase=True)

    name = mdf.save(filename, overwrite=True)

    mdf.close()


if __name__ == "__main__":
    #    generate_test_file("3.30")
    #    generate_test_file("4.10")
    generate_arrays_test_file(r"D:\TMP")



import unittest

import numpy as np

from asammdf.blocks import bus_logging_utils as blu


class TestCANTP(unittest.TestCase):
    tempdir = None

    payload = np.vstack(
        [
            np.frombuffer(b"\x10\x0b\x52\x49\x47\x20\x20\x39", dtype="uint8"),  # Initial part
            np.frombuffer(b"\x30\xff\x00\x4c\x40\x00\xd5\x54", dtype="uint8"),  # Flow control
            np.frombuffer(b"\x21\x30\x30\x30\x38\x33\x00\x00", dtype="uint8"),  # Final (second) part
            np.frombuffer(b"\x10\x0b\x52\x49\x47\x20\x20\x39", dtype="uint8"),  # Initial part of next frame...
        ]
    )
    ts = np.array([0.112, 0.113, 0.116, 0.201])

    def test_merge_cantp(self):

        merged, t = blu.merge_cantp(TestCANTP.payload, TestCANTP.ts)

        assert merged.shape == (1, 11)
        assert merged[0, -1] == 0x33
        assert t.shape == (1,)
        assert t[0] == 0.116
